import mymath;
import myio;
import texture;
import log;
import rc;
import camera;
import fpstimer;
import letter;
import vector;
import fpmgr;
import srv;
import readygo;

import SDL;
import SDL_pad;
import SDL_ttf;
import opengl;
import openglu;
import glee;
import gleeFuncDef;

import std.string;
import std.math;
import std.stdio;
import std.file;
import std.conv;

Texture tex;

extern (C) int red_x;
extern (C) int red_y;
extern (C) int audio_level;
extern (C) void start_capture_thread();
extern (C) void finish_capture_thread();

abstract class Pbg {
    static Vec3 mpos;
    float r, g, b;
    Vec3 pos;
    Vec3 vel;
    float life;

    // from NeHe19
    static float colors[12][3]=           // Rainbow Of Colors
    [
        [1.0f,0.5f,0.5f],[0.75f,1.0f,0.5f],[0.5f,0.5f,1.0f],[1.0f,0.75f,0.5f],
        [0.5f,1.0f,0.5f],[0.5f,1.0f,0.75f],[0.5f,1.0f,1.0f],[0.5f,0.75f,1.0f],
        [1.0f,1.0f,0.5f],[0.75f,0.5f,1.0f],[1.0f,0.5f,1.0f],[1.0f,0.5f,0.75f]
        ];

    this() { life = 1; }
    this(Vec3 p0) {
        this();
        pos = p0;
    }
    this(Vec3 p0, Vec3 v0) {
        this(p0);
        vel = v0;
    }
    this(float r0, float g0, float b0) {
        r = r0;
        g = g0;
        b = b0;
    }
    this(Vec3 p0, Vec3 v0, float r0, float g0, float b0) {
        this(p0, v0);
        r = r0;
        g = g0;
        b = b0;
    }

    abstract void proc();
    abstract void draw();
    void interact(Pbg p) {}

    float x() { return pos.x; }
    float y() { return pos.y; }
    float z() { return pos.z; }
    float x(float v) { return pos.x = v; }
    float y(float v) { return pos.y = v; }
    float z(float v) { return pos.z = v; }
    float vx() { return vel.x; }
    float vy() { return vel.y; }
    float vz() { return vel.z; }
    float vx(float v) { return vel.x = v; }
    float vy(float v) { return vel.y = v; }
    float vz(float v) { return vel.z = v; }
}

class Proton : Pbg {
    int nearCnt;

    this() {
        float[3] c = colors[rnd(12)];
        r = c[0];
        g = c[1];
        b = c[2];
        for (int i = 0; i < 3; i++) {
            pos[i] = rnd(mpos[i]);
            vel[i] = rnd(mpos[i]*0.02)-mpos[i]*0.01;
        }
    }
    void proc() {
        float vm = 5;
        foreach (inout float p, int i; pos) {
            if (vel[i] > vm) vel[i] = vm;
            else if (vel[i] < -vm) vel[i] = -vm;
            p += vel[i];
            if (p < 0) {
                p = 0;
                vel[i] = -vel[i];
                p += vel[i];
            }
            else if (pos[i] >= mpos[i]) {
                p = mpos[i];
                vel[i] = -vel[i];
                p += vel[i];
            }
        }

        float rs() { return rnd(2.0)-1.0; }
        Pbg pbg = new Powder(pos,
                             Vec3.create(-vx+rs(), -vy+rs(), -vz+rs()),
                             r, g, b);
        Main.powderMgr.add(pbg);

        nearCnt = 0;
    }
    void draw() {
        glEnable(GL_TEXTURE_2D);
        tex.bind();

        glPushMatrix();
        glTranslatef(x, y, z);

        glColor4f(r, g, b, 0.7);

        GLfloat[] texcoord = tex.texcoord;

//        float tx = 32 + ((vx > 0) ? 0 : -vx) * 10;
//        float ty = 32 + ((vy > 0) ? 0 : -vy) * 10;
        float tx = 64;
        float ty = 64;
        glBegin(GL_QUADS);
        glTexCoord3f(texcoord[0], texcoord[1], 0);
        glVertex3f(-tx, -ty, 0);
        glTexCoord3f(texcoord[0], texcoord[3], 0);
        glVertex3f(-tx, ty, 0);
        glTexCoord3f(texcoord[2], texcoord[3], 0);
        glVertex3f(tx, ty, 0);
        glTexCoord3f(texcoord[2], texcoord[1], 0);
        glVertex3f(tx, -ty, 0);
        glEnd();

        glPopMatrix();
    }
    void interact(Pbg p) {
        float r = 1;
        Vec3 dv = p.pos - pos;
        float d = 1.0 / dv.length();
        if (d < 0.1) {
            vel += r * dv * d;
            p.vel -= r * dv * d;
        }
        else {
            r *= 0.1;
            vel -= r * dv * d;
            p.vel += r * dv * d;

            nearCnt++;
            if (nearCnt > 3 && rnd(Thunder.genRatio) == 0) {
                float vr = 3;
                Pbg t = new Thunder(pos, vel*vr, r,g,b);
                Main.thunderMgr.add(t);
            }
        }
    }
}

class Card : Pbg {
    this() {
        r = 255;
        g = 0;
        b = 0;
        for (int i = 0; i < 3; i++) {
            pos[i] = 0;
            vel[i] = 0;
        }
    }
    void proc() {
        float rs() { return rnd(2.0)-1.0; }
        Pbg pbg = new Powder(pos,
                             Vec3.create(-vx+rs(), -vy+rs(), -vz+rs()),
                             r, g, b);
        Main.powderMgr.add(pbg);
    }
    void draw() {
        glEnable(GL_TEXTURE_2D);
        tex.bind();

        glPushMatrix();
        glTranslatef(x, y, z);

        glColor4f(r, g, b, 0.7);

        GLfloat[] texcoord = tex.texcoord;

//        float tx = 32 + ((vx > 0) ? 0 : -vx) * 10;
//        float ty = 32 + ((vy > 0) ? 0 : -vy) * 10;
        float tx = 16;
        float ty = 16;
        glBegin(GL_QUADS);
        glTexCoord3f(texcoord[0], texcoord[1], 0);
        glVertex3f(-tx, -ty, 0);
        glTexCoord3f(texcoord[0], texcoord[3], 0);
        glVertex3f(-tx, ty, 0);
        glTexCoord3f(texcoord[2], texcoord[3], 0);
        glVertex3f(tx, ty, 0);
        glTexCoord3f(texcoord[2], texcoord[1], 0);
        glVertex3f(tx, -ty, 0);
        glEnd();

        glPopMatrix();
    }
}

class TTFProton : Pbg {
    bool has_goal;
    Vec3 goal;
    int nearCnt;
    this() {
        float[3] c = colors[rnd(12)];
        r = c[0];
        g = c[1];
        b = c[2];
        for (int i = 0; i < 3; i++) {
            pos[i] = rnd(mpos[i]);
            vel[i] = rnd(mpos[i]*0.02)-mpos[i]*0.01;
        }
        goal.x = 0;
        goal.y = 0;
        goal.z = 0;
        has_goal = 0;
    }

    void setGoal(float x0, float y0) {
        goal.x = x0;
        goal.y = y0;
        goal.z = 0;
        has_goal = true;
    }

    void removeGoal() {
        has_goal = false;
        for (int i = 0; i < 3; i++) {
            vel[i] = rnd(mpos[i]*0.02)-mpos[i]*0.01;
        }
    }

    void proc() {
        if (has_goal) {
            vel *= 0.5;

            Vec3 dv = goal - pos;
            float d = 1.0 / dv.length();
            vel += dv * d * 5;
        }
        else {
            //vel *= 0.99;
        }
        if (red_x >= 0 && red_y >= 0) {
            Vec3 dv = Vec3.create((320-red_x) / 320.0 * 240.0 + 120,
                                  red_y / 240.0 * 240.0 + 120,
                                  0) - pos;
            float d = 1.0 / dv.length();
            vel += dv * d * d * 10;
        }

        float vm = 5;
        foreach (inout float p, int i; pos) {
            if (vel[i] > vm) vel[i] = vm;
            else if (vel[i] < -vm) vel[i] = -vm;
            p += vel[i];
            if (p < 0) {
                p = 0;
                vel[i] = -vel[i];
                p += vel[i];
            }
            else if (pos[i] >= mpos[i]) {
                p = mpos[i];
                vel[i] = -vel[i];
                p += vel[i];
            }
        }

        float ratio = 2;
        /*
        if (audio_level > 500000) {
            ratio *= 500000.0 / audio_level;
            if (ratio < 1) ratio = 1;
        }
        */
        if (rnd(ratio) < 1) {
            float rs() { return rnd(2.0)-1.0; }
            Pbg pbg = new Powder(pos,
                                 Vec3.create(-vx+rs(), -vy+rs(), -vz+rs()),
                                 r, g, b, 2);
            Main.powderMgr.add(pbg);
        }

        nearCnt = 0;
    }
    void draw() {
        glEnable(GL_TEXTURE_2D);
        tex.bind();

        glPushMatrix();
        glTranslatef(x, y, z);

        glColor4f(r, g, b, 0.7);

        GLfloat[] texcoord = tex.texcoord;

//        float tx = 32 + ((vx > 0) ? 0 : -vx) * 10;
//        float ty = 32 + ((vy > 0) ? 0 : -vy) * 10;
        float tx = 6;
        float ty = 6;
        glBegin(GL_QUADS);
        glTexCoord3f(texcoord[0], texcoord[1], 0);
        glVertex3f(-tx, -ty, 0);
        glTexCoord3f(texcoord[0], texcoord[3], 0);
        glVertex3f(-tx, ty, 0);
        glTexCoord3f(texcoord[2], texcoord[3], 0);
        glVertex3f(tx, ty, 0);
        glTexCoord3f(texcoord[2], texcoord[1], 0);
        glVertex3f(tx, -ty, 0);
        glEnd();

        glPopMatrix();
    }
    void interact(Pbg p0) {
        TTFProton p = cast(TTFProton)p0;
        if (!p) return;

        float r = 1;
        Vec3 dv = p.pos - pos;
        float d = 1.0 / dv.length();
        //float d = 1.0 / dv.length() * 0.01;

        if (d >= 0.01) {
            r *= -0.1;
        }
        else {
            nearCnt++;
            int ratio = Thunder.genRatio;
            if (audio_level > 5000000) {
                ratio *= 500000.0 / audio_level;
                if (ratio < 10) ratio = 10;
            }
            if (nearCnt > 3 && rnd(ratio) == 0) {
                float vr = 3;
                Pbg t = new Thunder(pos, vel*vr, r,g,b);
                Main.thunderMgr.add(t);
            }
        }

        if (!p.has_goal) {
            vel += r * dv * d;
        }
        if (!has_goal) {
            p.vel -= r * dv * d;
        }
    }
}

class Powder : Pbg {
    int sz;
    this(Vec3 p0, Vec3 v0, float r0, float g0, float b0, int sz0 = 8) {
        super(p0, v0, r0, g0, b0);
        sz = sz0;
    }
    void proc() {
        float vm = 1;
        float va = 0.2;
        foreach (inout float p, int i; pos) {
            if (vel[i] > vm) vel._[i] -= va;
            else if (vel[i] < -vm) vel._[i] += va;
            p += vel[i];
            if (p < 0) {
                p = 0;
                life = 0;
            }
            else if (p >= mpos[i]) {
                p = mpos[i];
                life = 0;
            }
        }
        life -= 0.1;
    }
    void draw() {
        glEnable(GL_TEXTURE_2D);
        tex.bind();

        glPushMatrix();
        glTranslatef(x, y, z);

        glColor4f(r, g, b, life);

        GLfloat[] texcoord = tex.texcoord;

        float tx = sz;
        float ty = sz;
        glBegin(GL_QUADS);
        glTexCoord3f(texcoord[0], texcoord[1], 0);
        glVertex3f(-tx, -ty, 0);
        glTexCoord3f(texcoord[0], texcoord[3], 0);
        glVertex3f(-tx, ty, 0);
        glTexCoord3f(texcoord[2], texcoord[3], 0);
        glVertex3f(tx, ty, 0);
        glTexCoord3f(texcoord[2], texcoord[1], 0);
        glVertex3f(tx, -ty, 0);
        glEnd();

        glPopMatrix();
    }
}

class Lines : Pbg {
    Lines pair;
    static const int HISTS = 50;
    float[HISTS] hx, hy, hz;
    float[HISTS] hr, hg, hb;
    float dr, dg, db;
    int t;
    this(Vec3 p0, Vec3 v0) {
        t = 0;
        float[3] c = colors[rnd(12)];
        r = c[0] * .3;
        g = c[1] * .3;
        b = c[2] * .3;
        super(p0, v0);
    }
    void setPair(Lines p0) {
        assert(pair is null);
        pair = p0;
    }
    void proc() {
        if (pair) {
            pair.proc();

            hr[t] = r;
            hg[t] = g;
            hb[t] = b;

            if (t == 0) {
                float[3] c = colors[rnd(12)];
                dr = (c[0]-r)/HISTS;
                dg = (c[1]-g)/HISTS;
                db = (c[2]-b)/HISTS;
            }

            r += dr;
            g += dg;
            b += db;
        }

        t++;
        if (t >= HISTS) t = 0;
        hx[t] = x;
        hy[t] = y;
        hz[t] = z;

        pos += vel;
        float rs() { return rnd(10.0) + 5; }
        for (int i = 0; i < 3; i++) {
            if (pos[i] < 0) {
                pos[i] = 0;
                vel[i] = rs();
            }
            else if (pos[i] >= mpos[i]) {
                pos[i] = mpos[i];
                vel[i] = -rs();
            }
        }
    }
    void draw() {
        if (!pair) return;

        glEnable(GL_TEXTURE_2D);
        tex.bind();
        GLfloat[] texcoord = tex.texcoord;

        glBegin(GL_LINES);
        for (int i = 0; i < HISTS; i++) {
            glColor4f(hr[i], hg[i], hb[i], 0.05);
            glTexCoord3f(texcoord[0], texcoord[1], 0);
            glVertex3f(hx[i], hy[i], hz[i]);
            glTexCoord3f(texcoord[2], texcoord[3], 0);
            glVertex3f(pair.hx[i], pair.hy[i], pair.hz[i]);
//            glVertex3f(hx[i], hy[i], 300);
//            glVertex3f(pair.hx[i], pair.hy[i], 300);
        }
        glEnd();
    }
}

class Thunder : Pbg {
    static int width;
    static float dlife;
    static int genRatio;

    float[] hx, hy, hz;
    int t;
    this(Vec3 p0, Vec3 v0, float r0, float g0, float b0) {
        t = 0;
        super(p0, v0, r0, g0, b0);
    }
    void proc() {
        t++;
        life -= dlife;
        hx ~= x;
        hy ~= y;
        hz ~= z;

        pos += vel;

        float rs() { return rnd(15)-7; }
        vel.x += rs;
        vel.y += rs;
        vel.z += rs;
    }
    void draw() {
        glEnable(GL_TEXTURE_2D);
        tex.bind();

        GLfloat[] tmp = tex.texcoord;
        GLfloat[4] texcoord;
        texcoord[0] = tmp[0]*0.9+tmp[2]*0.1;
        texcoord[1] = tmp[1]*0.66+tmp[3]*0.33;
        texcoord[2] = tmp[2]*0.9+tmp[0]*0.1;
        texcoord[3] = tmp[3]*0.66+tmp[1]*0.33;

        glColor4f(r, g, b, life);
        glBegin(GL_QUADS);
        float rs() { return rnd(width) - width/2; }
        for (int i = 0; i < t-1; i++) {
            glTexCoord3f(texcoord[0], texcoord[1], 0);
            glVertex3f(hx[i], hy[i], hz[i]);
            glTexCoord3f(texcoord[2], texcoord[1], 0);
            glVertex3f(hx[i]+rs(), hy[i]+rs(), hz[i]);
            glTexCoord3f(texcoord[2], texcoord[3], 0);
            glVertex3f(hx[i+1]+rs(), hy[i+1]+rs(), hz[i+1]);
            glTexCoord3f(texcoord[0], texcoord[3], 0);
            glVertex3f(hx[i+1], hy[i+1], hz[i+1]);
        }
        glEnd();
    }
}

class PolyLine : Pbg {
    Vec3[] hp, fp, gp;
    static float theta;
    int t;
    static float width;
    static float intensity;
    static int useTexture;
    this(Vec3 p0, Vec3 v0, float r0, float g0, float b0) {
        t = 0;
        super(p0, v0, r0, g0, b0);
    }
    void proc() {
        t++;
        life -= 0.02;
        hp ~= pos;

        static Vec3 c = { x:0, y:0, z:-1 };

        if (t > 1) {
            int tp = t - 3;
            if (tp < 0) tp = 0;
            Vec3 zv = c-pos;
            zv.normalize();
            Vec3 tv = hp[t-1]-hp[tp];
            tv.normalize();
            Vec3 dv = width * (tv.outerProduct(zv));
            fp ~= hp[t-2] + dv;
            gp ~= hp[t-2] - dv;
        }

        pos += vel;

        float co = cos(theta);
        float si = sin(theta);
        float vtmp = co*vx + si*vy;
        vy = -si*vx + co*vy;
        vx = vtmp;
    }
    void draw() {
        if (t < 3) return;

        if (useTexture) {
            glEnable(GL_TEXTURE_2D);
            tex.bind();
        }
        else {
            glDisable(GL_TEXTURE_2D);
        }

        GLfloat[] tmp = tex.texcoord;
        GLfloat[4] texcoord;
        texcoord[0] = tmp[0]*0.9+tmp[2]*0.1;
        texcoord[1] = tmp[1]*0.66+tmp[3]*0.33;
        texcoord[2] = tmp[2]*0.9+tmp[0]*0.1;
        texcoord[3] = tmp[3]*0.66+tmp[1]*0.33;

        glColor4f(r, g, b, life*intensity);
        glBegin(GL_QUADS);
        for (int i = 0; i < fp.length-1; i++) {
            glTexCoord3f(texcoord[0], texcoord[1], 0);
            glVertex3f(fp[i].x, fp[i].y, hp[i].z);
            glTexCoord3f(texcoord[2], texcoord[1], 0);
            glVertex3f(gp[i].x, gp[i].y, hp[i].z);
            glTexCoord3f(texcoord[2], texcoord[3], 0);
            glVertex3f(gp[i+1].x, gp[i+1].y, hp[i+1].z);
            glTexCoord3f(texcoord[0], texcoord[3], 0);
            glVertex3f(fp[i+1].x, fp[i+1].y, hp[i+1].z);
        }
        glEnd();
    }
}

class Simple : Pbg {
    this(Vec3 p0, Vec3 v0, float r0, float g0, float b0) {
        super(p0, v0, r0, g0, b0);
    }
    void proc() {
        foreach (inout float p, int i; pos) {
            p += vel[i];
/*
            if (p < 0) {
                p = 0;
                life = 0;
            }
            else if (p >= mpos[i]) {
                p = mpos[i];
                life = 0;
            }
*/
            if (z < -500) {
                life = 0;
            }
        }
        life -= 0.01;
    }
    void draw() {
        glEnable(GL_TEXTURE_2D);
        tex.bind();

//        if (life > 0.9) return;

        glPushMatrix();
        glTranslatef(x, y, z);

        glColor4f(r, g, b, life*0.5);

        GLfloat[] texcoord = tex.texcoord;

        float tx = 16;
        float ty = 16;
        glBegin(GL_QUADS);
        glTexCoord3f(texcoord[0], texcoord[1], 0);
        glVertex3f(-tx, -ty, 0);
        glTexCoord3f(texcoord[0], texcoord[3], 0);
        glVertex3f(-tx, ty, 0);
        glTexCoord3f(texcoord[2], texcoord[3], 0);
        glVertex3f(tx, ty, 0);
        glTexCoord3f(texcoord[2], texcoord[1], 0);
        glVertex3f(tx, -ty, 0);
        glEnd();

        glPopMatrix();
    }
}

class ParticleMgr {
    Pbg[][] particles;
    Pbg[][] nparticles;
    int num;
    int dx, dy, dz;
    float mx, my, mz;
    float fx, fy, fz;

    int particleNum() {
        int ret = 0;
        foreach (Pbg[] ps; particles) {
            ret += ps.length;
        }
        return ret;
    }

    int getZone(Pbg p) {
        int z = cast(int)(p.x/fx)+dx*(cast(int)(p.y/fy)+dy*cast(int)(p.z/fz));
        if (z < 0) z = 0;
        else if (z >= nparticles.length) z = nparticles.length-1;
        return z;
    }

    void proc() {
        foreach (int zone, Pbg[] ps; particles) {
            foreach (int i, Pbg p1; ps) {
                for (int j = i+1; j < ps.length; j++) {
                    p1.interact(ps[j]);
                }
            }
        }

        foreach (inout Pbg[] p; nparticles) p = null;
        foreach (int zone, Pbg[] ps; particles) {
            for (int i = 0; i < ps.length; i++) {
                Pbg p = ps[i];
                p.proc();

                if (p.life > 0) {
                    int z = getZone(p);
                    nparticles[z] ~= p;
                }
            }
        }
        foreach (int i, Pbg[] p; nparticles) particles[i] = p;
    }
    void draw() {
        foreach (Pbg[] ps; particles) {
            foreach (Pbg p; ps) {
                p.draw();
            }
        }
    }

    void add(Pbg pbg) {
        int z = getZone(pbg);
        particles[z] ~= pbg;
    }

    this(int dx0, int dy0, int dz0, float mx0, float my0, float mz0) {
        dx = dx0;
        dy = dy0;
        dz = dz0;
        mx = mx0;
        my = my0;
        mz = mz0;
        fx = mx / dx;
        fy = my / dy;
        fz = mz / dz;

        nparticles.length = particles.length = dx*dy*dz;
    }
}

Main_ Main;

class Main_ {
    static this() {
        Main = new Main_();
    }

    Fpmgr fpmgr_;

    int screenWidth_, screenHeight_;

    ParticleMgr protonMgr;
    ParticleMgr powderMgr;
    ParticleMgr linesMgr;
    ParticleMgr thunderMgr;
    ParticleMgr explodeMgr;
    ParticleMgr simpleMgr;
    ParticleMgr ttfMgr;
    ParticleMgr[] mgrs;

    FpsTimer timer;

    Rc rc;

    TTF_Font* ttf;
    TTF_Font* ttf_big;
    Srv srv;
    int numFreeTTFProton;

    Camera camera_;

    const(char[])texts[];
    ReadyGo textsTex;

    Pbg redCard;

    float red_x_hist[];
    int red_x_index;
    int last_red_move_turn;

    void resetTextTexture() {
        if (textsTex) {
            textsTex.finish();
        }
    }

    void showTextTexture() {
        textsTex = null;

        if (!texts) return;

        SDL_Color col;
        col.r = 255;
        col.g = 255;
        col.b = 255;
        SDL_Surface* surfs[];
        int maxw = 0;
        int h = 0;
        foreach (txt; texts) {
            SDL_Surface* s = TTF_RenderUTF8_Solid(ttf_big, toStringz(txt), col);
            maxw = s.w > maxw ? s.w : maxw;
            h += s.h + (surfs.length ? 20 : 0);
            surfs ~= s;
        }
        SDL_Surface* cs =
            SDL_CreateRGBSurface(SDL_SWSURFACE, maxw, h, 32, 0,0,0,0);
        int y = 0;
        foreach (i, s; surfs) {
            SDL_Rect r;
            r.x = 0;
            r.y = y;
            y += s.h + 20;
            SDL_BlitSurface(s, null, cs, &r);
            SDL_FreeSurface(s);
        }
        textsTex = new ReadyGo(new Texture(cs));
        SDL_FreeSurface(cs);
    }

    TTFProton getRandomTTFProton() {
        int r = rnd(ttfMgr.particleNum);
        for (int i = 0; i < ttfMgr.particles.length; i++) {
            if (r < ttfMgr.particles[i].length) {
                return cast(TTFProton)ttfMgr.particles[i][r];
            }
            r -= ttfMgr.particles[i].length;
        }
        assert(false);
    }

    TTFProton acquireTTFProton() {
        if (numFreeTTFProton == 0) {
            return getRandomTTFProton();
        }
        while (true) {
            TTFProton tp = getRandomTTFProton();
            if (!tp.has_goal) {
                numFreeTTFProton--;
                return tp;
            }
            tp = getRandomTTFProton();
        }
        assert(false);
    }

    void releaseAllTTFProton() {
        resetTextTexture;

        texts = null;

        foreach (tpa; ttfMgr.particles) {
            foreach (pbg; tpa) {
                TTFProton tp = cast(TTFProton)pbg;
                tp.removeGoal();
            }
        }
        numFreeTTFProton = ttfMgr.particleNum;
    }

    void setText(const char[] txt, int bx, int by) {
        texts ~= txt;

        SDL_Color col;
        col.r = 255;
        col.g = 255;
        col.b = 255;
        SDL_Surface* s = TTF_RenderUTF8_Solid(ttf, toStringz(txt), col);

        Log.info(format("%d %d %d", s.w, s.h, s.format.BytesPerPixel));
        int w = s.w > 80 ? 80 : s.w;
        for (int y = 0; y < s.h; y++) {
            for (int x = 0; x < w; x++) {
                Uint8 p = *(cast(Uint8*)s.pixels + x + y * s.pitch);
                if (p) {
                    TTFProton pbg = acquireTTFProton();
                    pbg.setGoal(x*5 + bx, y*5 + by);
                }
            }
        }

        SDL_FreeSurface(s);
    }

	void initSDL() {
		Log.info("init SDL");

		int ret = SDL_Init(SDL_INIT_VIDEO|SDL_INIT_TIMER|SDL_INIT_NOPARACHUTE);
		if (ret != 0) Log.errorSDL("cannot init SDL");

		ret = SDL_InitSubSystem(SDL_INIT_JOYSTICK);
		if (ret != 0) {
			Log.warnSDL("cannot init joystick");
		}

        TTF_Init();

        ttf = TTF_OpenFont(toStringz(`sazanami-gothic.ttf`), 10);
        ttf_big = TTF_OpenFont(toStringz(`sazanami-gothic.ttf`), 20);

/*
  useAudio_ = true;
  ret = SDL_InitSubSystem(SDL_INIT_AUDIO);
  if (ret != 0) {
  Log.warnSDL("cannot init audio");
  useAudio_ = false;
  }
*/

		SDL_VideoInfo* info = SDL_GetVideoInfo();

        int bpp = rc.getInt("screenBpp");
        char[] bppMsg;

		if (bpp == -1) {
			bpp = info.vfmt.BitsPerPixel;
			Log.info("using the \"best\" bpp: " ~ tostr(bpp));
		}
		else if (bpp == 0) {
			bpp = info.vfmt.BitsPerPixel;
			Log.info("using current bpp");
		}
		else {
			Log.info("using user set bpp: " ~ tostr(bpp));
		}

		SDL_GL_SetAttribute( SDL_GL_RED_SIZE, 5 );  
		SDL_GL_SetAttribute( SDL_GL_GREEN_SIZE, 5 );
		SDL_GL_SetAttribute( SDL_GL_BLUE_SIZE, 5 );  
		SDL_GL_SetAttribute( SDL_GL_DEPTH_SIZE, 16 );
		SDL_GL_SetAttribute( SDL_GL_DOUBLEBUFFER, 1 );

		screenWidth_ = rc.getInt("screenWidth");
		screenHeight_ = rc.getInt("screenHeight");
		int screenFull = rc.getInt("screenFull");

		int flags = SDL_OPENGL;

		if (screenFull) {
			flags |= SDL_FULLSCREEN;
		}

        Log.info("set video mode " ~
                 "size = " ~ tostr(screenWidth_) ~
                 "x" ~ tostr(screenHeight_) ~
                 ", bpp = " ~ tostr(bpp) ~
                 ", flags = " ~ tostr(flags));

		if (SDL_SetVideoMode(screenWidth_, screenHeight_,
							 bpp, flags) == null)
		{
			Log.errorSDL("cannot create screen");
		}

		info = SDL_GetVideoInfo();
		Log.info("set bpp: " ~ tostr(cast(int)info.vfmt.BitsPerPixel));

        SDL_WM_SetCaption(toStringz(`pbg`), null);

        Log.info("pad init");
        Pad_Init();

        Pad_LoadDefault(toStringz("padrc"));
	}

	void initGL() {
        float bgr = rc.getFloat("clearColR");
        float bgg = rc.getFloat("clearColG");
        float bgb = rc.getFloat("clearColB");
        if (rc.getInt("reverseColor")) {
            /*
            glClearColor(1-bgr, 1-bgg, 1-bgb, 0);

            glBlendEquationEXT(GL_FUNC_REVERSE_SUBTRACT_EXT);
            glBlendFunc(GL_SRC_ALPHA, GL_ONE);
            */
        }
        else {
            glClearColor(bgr, bgg, bgb, 0);
        }
        {
            GLeeInit();

            /*
            uint vshader;
            glEnable(GL_VERTEX_PROGRAM_ARB);
            glGenProgramsARB(1, &vshader);
            glBindProgramARB(GL_VERTEX_PROGRAM_ARB, vshader);
            glProgramStringARB(GL_VERTEX_PROGRAM_ARB,
                               GL_PROGRAM_FORMAT_ASCII_ARB,
                               vprog.length, vprog.ptr);
            printf("%s\n", gluErrorString(glGetError()));
            */

            fpmgr_.load();
        }

		glViewport(0, 0, screenWidth_, screenHeight_);

		glMatrixMode(GL_PROJECTION);
		glLoadIdentity();
		float asp = cast(float)screenWidth_ / screenHeight_;
		gluPerspective(45.0, asp, 0.1, 1000);

		glEnable(GL_BLEND);
		glBlendFunc(GL_SRC_ALPHA, GL_ONE);
//		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
//		glBlendFunc(GL_ONE, GL_ZERO);

		glMatrixMode(GL_MODELVIEW);
        glLoadIdentity();

		// from PARSEC47
		glDisable(GL_LIGHTING);
		glDisable(GL_CULL_FACE);
		glDisable(GL_DEPTH);
		glDisable(GL_DEPTH_TEST);
        glDisable(GL_TEXTURE_2D);

        glLineWidth(rc.getInt("bgLineWidth"));
	}

    void handleOps() {
        char[][] ops = srv.retrieveOps();
        foreach (opstr; ops) {
            string[] op = std.string.split(cast(string)opstr, "\t");
            if (op.length == 0) {
                Log.info(`No operation: ` ~ opstr);
                continue;
            }
            string opc = op[0];
            if (opc == `release`) {
                releaseAllTTFProton();
                Log.info("release");
            }
            else if (opc == `text` && op.length == 4) {
                string t = op[1];
                int x = toInt(op[2]);
                int y = toInt(op[3]);
                setText(t, x, y);
                Log.info(format("text: %d %d", x, y));
            }
            /*
            else if (opc == `acquire`) {
                TTFProton tp = acquireTTFProton();
            }
            else if (opc == `move`) {
                //int 
            }
            */
            else {
                Log.warn(`Unknown operation: ` ~ opstr);
            }
        }
    }

    int main(char[][] args) {
        Log.init();

        rc = new Rc();
        rc.addResourceFile(`pbg.rc`);
        fpmgr_ = new Fpmgr();
        initSDL();
        initGL();
        initLetter();

        srv = new Srv;
        srv.start();

        int enableProton = rc.getInt("enableProton");
        int enableLines = rc.getInt("enableLines");
        int enableThunder = rc.getInt("enableThunder");
        int enableTrafficLight = rc.getInt("enableTrafficLight");
        int enableFlame = rc.getInt("enableFlame");
        int enableSystemInfo = rc.getInt("enableSystemInfo");

        Thunder.width = rc.getInt("thunderWidth");
        Thunder.dlife = 1.0 / rc.getInt("thunderLifetime");
        Thunder.genRatio = rc.getInt("thunderGenRatio");

        PolyLine.width = rc.getFloat("trafficLightWidth");
        PolyLine.intensity = rc.getFloat("trafficLightIntensity");
        PolyLine.useTexture = rc.getInt("trafficUseTexture");
        int trafficLightNum = rc.getInt("trafficLightNum");

        int flamePercentage = rc.getInt("flamePercentage");

        tex = new Texture(`particle.bmp`);

        Pbg.mpos.x = 480;
        Pbg.mpos.y = 480;
        Pbg.mpos.z = 300;

        redCard = new Card();
        red_x_hist = new float[10];
        for (int i = 0; i < red_x_hist.length; i++) {
            red_x_hist[i] = -1;
        }

        protonMgr = new ParticleMgr(4, 4, 4, 480, 480, 300);
        for (int i = 0; i < 100; i++) {
            Pbg pbg = new Proton();
            protonMgr.add(pbg);
        }

        Main.powderMgr = new ParticleMgr(4, 4, 4, 480, 480, 300);

        linesMgr = new ParticleMgr(1, 1, 1, 480, 480, 300);
        float rs() { return rnd(10.0) + 5; }
        for (int i = 0; i < 20; i++) {
            Lines l1 = new Lines(Vec3.zero(), Vec3.generate(&rs));
            Lines l2 = new Lines(Vec3.zero(), Vec3.generate(&rs));
            l1.setPair(l2);
            linesMgr.add(l1);
        }

        thunderMgr = new ParticleMgr(1, 1, 1, 480, 480, 300);

        explodeMgr = new ParticleMgr(1, 1, 1, 480, 480, 300);

        simpleMgr = new ParticleMgr(1, 1, 1, 480, 480, 300);

        if (enableProton) {
            //mgrs ~= protonMgr;
            mgrs ~= powderMgr;
        }
        if (enableLines) mgrs ~= linesMgr;
        if (enableThunder) mgrs ~= thunderMgr;
        if (enableTrafficLight) mgrs ~= explodeMgr;
        if (enableFlame) mgrs ~= simpleMgr;

        ttfMgr = new ParticleMgr(3, 3, 3, 480, 480, 300);
        for (int i = 0; i < 300; i++) {
            numFreeTTFProton++;
            ttfMgr.add(new TTFProton());
        }
        //setText(`日本語`, 100, 100);
        mgrs ~= ttfMgr;

        int cameraId = rc.getInt("cameraId");
        if (cameraId == -1) {
            camera_ = new Camera();
            camera_.change(0);
        }
        else {
            camera_ = new Camera(cameraId);
        }

        timer = new FpsTimer(true);
		int frameSkip = rc.getInt("frameSkip");
        if (args.length >= 2) frameSkip = 0;
        if (frameSkip) {
            int fps = rc.getInt("fps");
            timer.setFps(fps);
        }
        else {
            timer.setFrameSkip(false);
        }

        Uint32 startTicks = SDL_GetTicks();

        int turn;
        for (turn = 0; true; turn++) {
            //printf("%d %d\n", red_x, red_y);
            srv.proc();
            handleOps();

            Pad_Pump(PAD_PUMP_EVERYTHING);
            if (Pad_GetButtonPushed(PAD_END_BUTTON)) {
                break;
            }
            if (Pad_GetButtonPushed(PadKey.PADK_PAGEUP)) {
                camera_.changePrev();
            }
            if (Pad_GetButtonPushed(PadKey.PADK_PAGEDOWN)) {
                camera_.changeNext();
            }
            if (Pad_GetButtonPushed(PadKey.PADK_t)) {
                if (textsTex) {
                    resetTextTexture();
                }
                else {
                    showTextTexture();
                }
            }
            if (Pad_GetButtonPushed(PAD_A_BUTTON)) {
                camera_.end();
            }
            if (Pad_GetButtonPushed(PAD_B_BUTTON)) {
                if (camera_.stop) Log.info("camera move");
                else Log.info("camera stop");
                camera_.stop = !camera_.stop;
            }
            if (Pad_GetButtonPushed(PAD_C_BUTTON)) {
                fpmgr_.copy();
            }
            if (Pad_GetButtonPushed(PAD_D_BUTTON)) {
                fpmgr_.reload();
            }
            if (Pad_GetButtonPushed(PAD_DOWN_BUTTON)) {
                fpmgr_.next();
                fpmgr_.reload();
            }
            if (Pad_GetButtonPushed(PAD_UP_BUTTON)) {
                fpmgr_.prev();
                fpmgr_.reload();
            }
            if (Pad_GetButtonPushed(PAD_RIGHT_BUTTON)) {
                srv.addEvent(`RIGHT`);
            }
            if (Pad_GetButtonPushed(PAD_LEFT_BUTTON)) {
                srv.addEvent(`LEFT`);
            }

            if (textsTex) {
                textsTex.proc();
                if (textsTex.done()) {
                    textsTex = null;
                }
            }

            // explodeMgr
            if (enableTrafficLight) {
                float a = pow(turn,1.5)/20;
                int t = (turn/10)%80;
                if (t > 60) t = t-80;
                else if (t > 40) t = -t+40;
                else if (t > 20) t = 40-t;

                static Vec3[] p,v;
                static bool inited = false;
                static float[3][] col;
                float rv() { return cast(float)(rnd(15))/10; }
                if (!inited) {
                    inited = true;

                    col.length = trafficLightNum;
                    for (int i = 0; i < trafficLightNum; i++) {
                        col[i][] = Pbg.colors[i];
                    }

                    p.length = trafficLightNum;
                    v.length = trafficLightNum;

                    for (int i = 0; i < trafficLightNum; i++) {
                        p[i].x = 240;
                        p[i].y = 240;
                        p[i].z = 200;
                        for (int j = 0; j < 2; j++) {
                            v[i][j] = rv();
                            if (rnd(2) == 0) v[i][j] = -v[i][j];
                        }
                    }
                }

                for (int i = 0; i < trafficLightNum; i++) {
                    p[i] += v[i];
                    for (int j = 0; j < 3; j++) {
                        if (p[i][j] < 0) {
                            v[i][j] = rv();
                        }
                        else if (p[i][j] > 480) {
                            v[i][j] = -rv();
                        }
                    }
                }

                PolyLine.theta = cast(float)(t)/80.0;
                for (int i = 0; i < trafficLightNum; i++) {
//                    if (rnd(3) == 0) continue;
                    Pbg pbg = new PolyLine(Vec3.create(p[i][0],p[i][1],200),
                                           Vec3.create(cos(a)*10,sin(a)*10,0),
                                           col[i][0],col[i][1],col[i][2]);
                    explodeMgr.add(pbg);
                }
//            }

            // simpleMgr
                if (enableFlame) {
                    for (int i = 0; i < trafficLightNum; i++) {
                        if (rnd(100) > flamePercentage) continue;
                        Pbg pbg = new Simple(Vec3.create(p[i][0],p[i][1],200),
                                             Vec3.create(rnd(2.0)-1,rnd(2.0)-1,-8),
//                                         1,1,1);
                                             col[i][0],col[i][1],col[i][2]);
                        simpleMgr.add(pbg);
                    }
                }
            }

            foreach (ParticleMgr m; mgrs) m.proc();

            if (!timer.shouldSkip()) {
                glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);
                glMatrixMode(GL_MODELVIEW);
                glLoadIdentity();

                float t = turn / 20.0;
                glProgramEnvParameter4dARB(
                    GL_FRAGMENT_PROGRAM_ARB, 0,
                    sin(t), cos(t), 0, 0);

                if (turn < 200 && cameraId == -1) {
                    gluLookAt(240, 240, -500, 240, 240, 0, 0, -1, 0);
                }
                else {
                    camera_.set();
                    glTranslatef(80, 0, 0);
                }

                foreach (ParticleMgr m; mgrs) m.draw();

                if (enableSystemInfo) drawSystemInfo();

                if (textsTex) {
                    textsTex.draw();
                }

                if (red_x >= 0 && red_y >= 0) {
                    redCard.x = (320-red_x) / 320.0 * 240.0 + 120;
                    redCard.y = red_y / 240.0 * 240.0 + 120;
                    redCard.draw();
                }
                if (red_x > 0 && red_x_hist[red_x_index] > 0) {
                    if (red_x > red_x_hist[red_x_index] + 30) {
                        Log.info(`left move detected`);
                        if (last_red_move_turn + 100 < turn &&
                            srv.polling) {
                            srv.addEvent(`LEFT`);
                        }
                        last_red_move_turn = turn;
                    }
                    if (red_x < red_x_hist[red_x_index] - 30) {
                        Log.info(`right move detected`);
                        if (last_red_move_turn + 100 < turn &&
                            srv.polling) {
                            srv.addEvent(`RIGHT`);
                        }
                        last_red_move_turn = turn;
                    }
                }
                ++red_x_index;
                if (red_x_index >= red_x_hist.length) {
                    red_x_index = 0;
                }
                red_x_hist[red_x_index] = red_x;
                // drawBox();

                SDL_GL_SwapBuffers();
            }
            timer.wait();
//            if (frameSkip) timer.wait();
//        SDL_Delay(16);
        }

        Uint32 endTicks = SDL_GetTicks();
        printf("%d FPS.\n", turn*1000/(endTicks-startTicks));

        srv.finish();

        TTF_Quit();

        SDL_Quit();

        return 0;
    }

    void drawSystemInfo() {
        glDisable(GL_FRAGMENT_PROGRAM_ARB);

        glLoadIdentity();
        gluLookAt(240, 240, -500, 240, 240, 0, 0, -1, 0);

        int x = -35;
        int y = 45;
        glDisable(GL_TEXTURE_2D);
        drawString(`FPS`, x,y, 10,12, 1,1,1);
        drawNum(timer.getTrueFps(), x+65,y, 10,12, 1,1,1);
        y += 15;
        //drawString(`FRAMES`, x,y, 10,12, 1,1,1);
        //drawNum(timer.getTurn(), x+65,y, 10,12, 1,1,1);
        drawString(`SECS`, x,y, 10,12, 1,1,1);
        drawNum(timer.getEllapsedSecond(), x+65,y, 10,12, 1,1,1);
        y += 15;
        drawString(`DROPS`, x,y, 10,12, 1,1,1);
        drawNum(timer.getDroppedTurn(), x+65,y, 10,12, 1,1,1);
/*
        y += 15;
        drawString(`SDROPS`, x,y, 10,12, 1,1,1);
        drawNum(timer.getSystemDroppedTurn(), x+65,y, 10,12, 1,1,1);
*/
        int cnt;
        foreach (ParticleMgr m; mgrs) {
            cnt += m.particleNum;
        }
        y += 15;
        drawString(`OBJS`, x,y, 10,12, 1,1,1);
        drawNum(cnt, x+65,y, 10,12, 1,1,1);
        y += 15;
        drawString(`PS#`, x,y, 10,12, 1,1,1);
        drawNum(fpmgr_.index, x+65,y, 10,12, 1,1,1);
        y += 15;
        drawString(`CAM#`, x,y, 10,12, 1,1,1);
        drawNum(camera_.currentId, x+65,y, 10,12, 1,1,1);

        glEnable(GL_FRAGMENT_PROGRAM_ARB);
    }

    void drawBox() {
/*
  glColor4f(1,1,1,0.8);
  glDisable(GL_TEXTURE_2D);
  glBegin(GL_LINE_LOOP);
  glVertex3f(0, 0, 0);
  glVertex3f(0, 480, 0);
  glVertex3f(480, 480, 0);
  glVertex3f(480, 0, 0);
  glEnd();
  glBegin(GL_LINE_LOOP);
  glVertex3f(0, 0, 300);
  glVertex3f(0, 480, 300);
  glVertex3f(480, 480, 300);
  glVertex3f(480, 0, 300);
  glEnd();
  glBegin(GL_LINES);
  glVertex3f(0, 0, 0);
  glVertex3f(0, 0, 300);
  glVertex3f(0, 480, 0);
  glVertex3f(0, 480, 300);
  glVertex3f(480, 480, 0);
  glVertex3f(480, 480, 300);
  glVertex3f(480, 0, 0);
  glVertex3f(480, 0, 300);
  glEnd();
*/
    }

    float targetX() {
        int n = 5;
        while (protonMgr.particles[n].length == 0) n++;
        return protonMgr.particles[n][0].x;
    }
    float targetY() {
        int n = 5;
        while (protonMgr.particles[n].length == 0) n++;
        return protonMgr.particles[n][0].y;
    }
}

int main(char[][] args) {
    start_capture_thread();
    int r = Main.main(args);
    finish_capture_thread();
    return r;
}

