private import std.conv;
private import std.cstream;
private import std.string;
private import log;
private import myio;

// for atof
private import std.c.stdlib;

class Rc {
public:
	this() {
	}

	void addDefault() {
	}
	void addResourceFile() {
		addResourceFile("data/kk.rc");
		addResourceFile("save/kk.rc");
	}
	void addResourceFile(string file) {
		auto File input;
		try {
			input = new File(file);
		}
		catch (Error e) {
			Log.error(e.toString());
		}

		if (input.eof()) {
			Log.error(file ~ ": file not found");
		}

		while (!input.eof()) {
			string line = cast(string)input.readLine();
			if (line.length == 0 || line[0] == '#') continue;
			string[] tok = split(line);
			if (tok.length < 2) {
				Log.warn(file ~  ": resource file broken. (" ~ line ~ ")");
			}

			int valIndex = (tok[1] == "=") ? 2 : 1;

			set(tok[0], join(tok[valIndex..tok.length], " "));
		}
	}
	void addArgument(char args[][]) {
		for (int i = 1; i < args.length; i++) {
			if (args[i] == "-w") {
				set("timerEnable", "0");
			}
			else {
				if (args[i] != "-h") {
					dout.writeLine(args[i] ~ ": invalid option");
				}
				dout.writeLine("usage: " ~ args[0] ~ " [options]");
				dout.writeLine("options:");
				dout.writeLine("  -w    \tnowait");
				dout.writeLine("  -m    \tnosound (yet)");
				dout.writeLine("  -l    \tlow resolution (yet)");
				dout.writeLine("  -b    \tno background (yet)");
				dout.writeLine("  -w    \twindow (yet)");
				dout.writeLine("  -h    \thelp");
				exit(1);
			} 
		}
	}

public:
	bit hasKey(string key) {
		return (rc_[key] !is null);
	}

	string get(string key) {
		if (!rc_[key]) {
			Log.warn(key ~ " is not in resource");
			return "";
		}
		return rc_[key];
	}
	int getInt(string key) {
		return toInt(get(key));
	}
	bit getBit(string key) {
		return toInt(get(key)) != 0;
	}
	float getFloat(string key) {
		return std.string.atof(get(key));
	}

	void set(string key, string val) {
		rc_[key] = val;
	}

	string[] keys() {
		return rc_.keys;
	}

private:
	string[string] rc_;

};

private import std.stream;

/*
unittest {
	void doAssert(bit b) {
		if (!b) {
			stdout.writeLine("assert failure");
		}
	}

	auto Rc rc = new Rc();
	rc.set("hoge", "huga");
	doAssert("huga" == rc.get("hoge"));
	doAssert(null == rc.get("huga"));
}
*/
